package de.ronnyfriedland.maven.plugin;

import java.io.File;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.Proxy;
import java.net.ProxySelector;
import java.net.SocketAddress;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.UUID;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;

import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;

import de.ronnyfriedland.maven.plugin.item.Item;
import de.ronnyfriedland.maven.plugin.item.StringArrayItem;
import de.ronnyfriedland.maven.plugin.item.StringItem;
import de.ronnyfriedland.maven.plugin.util.LdapUtils;
import de.ronnyfriedland.maven.plugin.value.ValueItem;

/**
 * Searches the given rootdn and checks if the given groupid and artifactid of
 * the source project already stored in ldap. If not the nodes are created added
 * to the parent (groupid).
 * 
 * <pre>
 *   &lt;configuration&gt;
 *     &lt;host&gt;localhost&lt;/host&gt;
 *     &lt;port&gt;10389&lt;/port&gt;
 *     &lt;user&gt;uid=admin,ou=system&lt;/user&gt;
 *     &lt;password&gt;secret&lt;/password&gt;
 *     &lt;rootdn&gt;ou=repository,ou=maven&lt;/rootdn&gt;
 *     &lt;ssl&gt;false&lt;/ssl&gt;
 *     &lt;proxyHost&gt;proxy&lt;/proxyHost&gt;
 *     &lt;proxyPort&gt;12345&lt;/proxyPort&gt;
 *    &lt;/configuration&gt;
 * </pre>
 * 
 * groupid: groupOfUniqueNames <br/>
 * artifactid: organizationalUnit, uidObject, labeledURIObject, extensibleObject
 * 
 * @author Ronny Friedland
 * 
 * @description Plugin to upload a successful maven build/artifact to an ldap
 *              repository.
 * @goal ldap
 * @phase install
 * 
 */
public class LdapArtifactMojo extends AbstractMojo {

    /**
     * Proxy selector implementation to access ldap host via proxy. Stores the
     * default proxy selector.
     * 
     * @author Ronny Friedland
     */
    class LdapProxySelector extends ProxySelector {
        private final ProxySelector defaultProxySelector;

        /**
         * Creates a new instance of {@link LdapProxySelector}.
         * 
         * @param defaultProxySelector
         *            the default proxy selector
         */
        public LdapProxySelector(ProxySelector defaultProxySelector) {
            this.defaultProxySelector = defaultProxySelector;
        }

        /**
         * Returns the default proxy selector
         * 
         * @return the default proxy selector
         */
        public ProxySelector getDefaultProxySelector() {
            return defaultProxySelector;
        }

        @Override
        public List<Proxy> select(URI uri) {
            List<Proxy> proxies = new ArrayList<Proxy>();
            if (null != proxyHost && null != proxyPort) {
                SocketAddress addr = new InetSocketAddress(proxyHost, proxyPort);
                Proxy proxy = new Proxy(Proxy.Type.SOCKS, addr);
                proxies.add(proxy);
            } else {
                proxies.add(Proxy.NO_PROXY);
            }
            return proxies;
        }

        @Override
        public void connectFailed(URI uri, SocketAddress sa, IOException ioe) {
            throw new RuntimeException(ioe);
        }
    }

    /**
     * The root dn.
     * 
     * @parameter expression="dc=example,dc=com"
     * @required
     */
    private String rootdn;

    /**
     * The target ldap host.
     * 
     * @parameter expression="localhost"
     * @required
     */
    private String host;

    /**
     * The target ldap port.
     * 
     * @parameter expression="389"
     * @required
     */
    private Integer port;

    /**
     * The user to connect to ldap host.
     * 
     * @parameter expression=""
     */
    private String user;

    /**
     * The password to connect to ldap host.
     * 
     * @parameter expression=""
     */
    private String password;

    /**
     * Flag if ldap connection uses ssl.
     * 
     * @parameter expression="false"
     */
    private boolean ssl;

    /**
     * The keystore to connect to ldap using ssl.
     * 
     * @parameter expression="keystore.jks"
     */
    private String keystore;

    /**
     * The keystore password.
     * 
     * @parameter expression="password"
     */
    private String keystorePassword;

    /**
     * The truststore to validate ssl server certificate.
     * 
     * @parameter expression="cacerts.jks"
     */
    private String truststore;

    /**
     * The truststore password.
     * 
     * @parameter expression="password"
     */
    private String truststorePassword;

    /**
     * The proxy host (optional)
     * 
     * @parameter
     */
    private String proxyHost;

    /**
     * The proxy host (optional)
     * 
     * @parameter
     */
    private Integer proxyPort;

    /**
     * @parameter default-value="${project}"
     * @readonly
     */
    protected MavenProject project;

    /**
     * @parameter expression="${localRepository}"
     * @readonly
     */
    protected ArtifactRepository local;

    /**
     * Creates a new instance of {@link LdapArtifactMojo}.
     */
    public LdapArtifactMojo() {
        super();
    }

    /**
     * Creates a new instance of {@link LdapArtifactMojo}.
     * 
     * @param rootdn
     *            the root dn
     * @param host
     *            the target host
     * @param port
     *            the target port
     * @param user
     *            the user to connect
     * @param password
     *            the password
     * @param ssl
     *            flag if connect uses ssl
     * @param keystore
     *            the keystore to use
     * @param keystorePassword
     *            the keystore password
     * @param truststore
     *            the truststore to use
     * @param truststorePassword
     *            the truststore password
     */
    public LdapArtifactMojo(final String rootdn, final String host, final Integer port, final String user,
            final String password, final boolean ssl, final String keystore, final String keystorePassword,
            final String truststore, final String truststorePassword) {
        this();
        this.rootdn = rootdn;
        this.host = host;
        this.port = port;
        this.user = user;
        this.password = password;
        this.ssl = ssl;
        this.keystore = keystore;
        this.keystorePassword = keystorePassword;
        this.truststore = truststore;
        this.truststorePassword = truststorePassword;
    }

    /**
     * {@inheritDoc}
     * 
     * @see org.apache.maven.plugin.Mojo#execute()
     */
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        String groupId = project.getGroupId();
        String artifactId = project.getArtifactId();
        String version = project.getVersion();

        LdapProxySelector ldapProxySelector = new LdapProxySelector(ProxySelector.getDefault());
        try {
            ProxySelector.setDefault(ldapProxySelector);
            Properties env = new Properties();
            env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
            if (ssl) {
                System.setProperty("javax.net.ssl.keyStore", keystore);
                System.setProperty("javax.net.ssl.keyStorePassword", keystorePassword);
                System.setProperty("javax.net.ssl.trustStore", truststore);
                System.setProperty("javax.net.ssl.trustStorePassword", truststorePassword);
                env.put(Context.PROVIDER_URL, "ldaps://" + host + ":" + port.intValue());
            } else {
                env.put(Context.PROVIDER_URL, "ldap://" + host + ":" + port.intValue());
            }
            env.put(Context.SECURITY_AUTHENTICATION, "simple");
            env.put(Context.SECURITY_PRINCIPAL, user);
            env.put(Context.SECURITY_CREDENTIALS, password);
            DirContext ctx = new InitialDirContext(env);

            try {
                checkGroupEntry(ctx, groupId);
                checkArtifactEntry(ctx, groupId, artifactId, version);
            } finally {
                ctx.close();
            }
        } catch (NamingException e) {
            getLog().error("Error accessing ldap server");
            throw new MojoExecutionException("Error accessing ldap server", e);
        } finally {
            ProxySelector.setDefault(ldapProxySelector.getDefaultProxySelector());
        }

    }

    private void checkGroupEntry(final DirContext ctx, final String groupId) throws NamingException {
        Map<String, Set<String>> results = LdapUtils.findMemberOfEntry(ctx, rootdn,
                LdapAttributes.UNIQUEMEMBER.getValue(),
                String.format("(%1$s=%2$s)", LdapAttributes.CN.getValue(), groupId));
        if (results.isEmpty()) {
            // create parent
            String[] parents = groupId.split("[.]");
            List<String> path = new ArrayList<String>();
            for (String parent : parents) {
                path.add(parent);
                Map<String, Set<String>> foundEntry = LdapUtils.findMemberOfEntry(ctx, rootdn,
                        LdapAttributes.UNIQUEMEMBER.getValue(),
                        String.format("(%1$s=%2$s)", LdapAttributes.CN.getValue(), parent));
                if (foundEntry.isEmpty()) {
                    // create
                    StringBuilder targetName = new StringBuilder();
                    for (int i = path.size() - 1; i >= 0; i--) {
                        targetName.append(String.format("%1$s=%2$s,", LdapAttributes.OU.getValue(), path.get(i)));
                    }
                    StringBuilder commonName = new StringBuilder();
                    for (int i = 0; i < path.size(); i++) {
                        commonName.append(path.get(i));
                        if (i < (path.size() - 1)) {
                            commonName.append(".");
                        }
                    }
                    createGroupEntry(ctx, parent, commonName.toString(),
                            String.format("%1$s%2$s", targetName.toString(), rootdn));
                }
            }
        }
    }

    private void checkArtifactEntry(final DirContext ctx, final String groupId, final String artifactId,
            final String version) throws NamingException {
        Map<String, Set<String>> results = LdapUtils.findMemberOfEntry(ctx, rootdn,
                LdapAttributes.UNIQUEMEMBER.getValue(),
                String.format("(%1$s=%2$s)", LdapAttributes.CN.getValue(), groupId));
        for (Map.Entry<String, Set<String>> result : results.entrySet()) {
            boolean exists = false;
            String parent = result.getKey() + "," + rootdn;
            String targetEntry = String.format("%1$s=%2$s,%3$s", LdapAttributes.OU.getValue(), artifactId, parent);
            String url = String.format("%1$s%7$s%2$s%7$s%3$s%7$s%4$s%7$s%5$s.%6$s", local.getUrl(), groupId.replaceAll(
                    "[.]", System.getProperty("path.separator")), artifactId, version, project.getBuild()
                    .getFinalName(), project.getPackaging(), File.pathSeparator);
            for (String member : result.getValue()) {
                if (member.equalsIgnoreCase(targetEntry)) {
                    // refresh entry
                    updateArtifactEntry(ctx, version, url, targetEntry);
                    exists = true;
                    break;
                }
            }
            if (!exists) {
                // create entry
                createArtifactEntry(ctx, parent, artifactId, version, url, targetEntry);
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void createGroupEntry(final DirContext ctx, final String entryName, final String commonName,
            final String targetName) throws NamingException {
        ValueItem<Item>[] values = new ValueItem[] {
                new ValueItem<StringArrayItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.OBJECTCLASS.getValue(),
                        new StringArrayItem("groupOfUniqueNames", "top")),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.UNIQUEMEMBER.getValue(),
                        new StringItem(targetName)),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.CN.getValue(), new StringItem(
                        commonName)),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.OU.getValue(), new StringItem(
                        entryName)) };
        // add entry
        LdapUtils.addEntry(ctx, targetName, values);
    }

    @SuppressWarnings("unchecked")
    private void createArtifactEntry(final DirContext ctx, final String parentName, final String entryName,
            final String version, final String url, final String targetName) throws NamingException {
        ValueItem<Item>[] values = new ValueItem[] {
                new ValueItem<StringArrayItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.OBJECTCLASS.getValue(),
                        new StringArrayItem("uidObject", "top", "organizationalUnit", "extensibleObject",
                                "labeledURIObject")),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.UID.getValue(), new StringItem(UUID
                        .randomUUID().toString())),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.OU.getValue(), new StringItem(
                        entryName)),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.LABELED_URI.getValue(),
                        new StringItem(url)),
                new ValueItem<StringItem>(DirContext.ADD_ATTRIBUTE, LdapAttributes.SERIALNUMBER.getValue(),
                        new StringItem(version)) };
        // add entry
        LdapUtils.addEntry(ctx, targetName, values);
        // modify parent
        LdapUtils.modifyMember(ctx, parentName, new ValueItem<Item>(DirContext.ADD_ATTRIBUTE,
                LdapAttributes.UNIQUEMEMBER.getValue(), new StringItem(targetName)));
    }

    @SuppressWarnings("unchecked")
    private void updateArtifactEntry(final DirContext ctx, final String version, final String url,
            final String targetName) throws NamingException {
        // modify entry
        ValueItem<Item>[] values = new ValueItem[] {
                new ValueItem<StringItem>(DirContext.REPLACE_ATTRIBUTE, LdapAttributes.SERIALNUMBER.getValue(),
                        new StringItem(version)),
                new ValueItem<StringItem>(DirContext.REPLACE_ATTRIBUTE, LdapAttributes.LABELED_URI.getValue(),
                        new StringItem(url)) };
        LdapUtils.modifyMember(ctx, targetName, values);
    }
}
