package de.ronnyfriedland.maven.plugin.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.ModificationItem;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;

import de.ronnyfriedland.maven.plugin.item.Item;
import de.ronnyfriedland.maven.plugin.item.StringArrayItem;
import de.ronnyfriedland.maven.plugin.item.StringItem;
import de.ronnyfriedland.maven.plugin.value.ValueItem;

/**
 * Utility class to access ldap.
 * 
 * @author Ronny Friedland
 */
public final class LdapUtils {

    /**
     * Searches an entry with the given filter. The method returns the given attributes.
     * 
     * @param ctx the directory context
     * @param rootDn the base dn of the search
     * @param returnAttribute defines the attribute to return - if <code>null</code>, all attributes are returned
     * @param filters the filter
     * @return all matching entries
     * @throws NamingException error searching entries
     */
    public static Map<String, Set<String>> findMemberOfEntry(final DirContext ctx, final String rootDn,
            final String returnAttribute, final String... filters) throws NamingException {
        Map<String, Set<String>> results = new HashMap<String, Set<String>>();
        SearchControls sc = new SearchControls();
        if (null != returnAttribute) {
            sc.setReturningAttributes(new String[] { returnAttribute });
        } else {
            sc.setReturningAttributes(new String[] { "*" });
        }
        sc.setSearchScope(SearchControls.SUBTREE_SCOPE);

        StringBuilder filterString = new StringBuilder();
        filterString.append("(&");
        if (filters.length == 0) {
            filterString.append("(").append(String.format("%1$s=%2$s", "*", "*")).append(")");
        } else {
            for (String filter : filters) {
                filterString.append("(").append(filter).append(")");
            }
        }
        filterString.append(")");

        NamingEnumeration<SearchResult> searchResults = ctx.search(rootDn, filterString.toString(), sc);
        try {
            while (searchResults.hasMoreElements()) {
                Set<String> members = new HashSet<String>();
                SearchResult searchResult = searchResults.nextElement();
                NamingEnumeration<? extends Attribute> attrs = searchResult.getAttributes().getAll();
                while (attrs.hasMoreElements()) {
                    NamingEnumeration<?> attributes = attrs.nextElement().getAll();
                    while (attributes.hasMoreElements()) {
                        String attribute = (String) attributes.nextElement();
                        members.add(attribute);
                    }
                }
                results.put(searchResult.getName(), members);
            }
        } finally {
            searchResults.close();
        }
        return results;
    }

    /**
     * Add entry to ldap structure.
     * 
     * @param ctx the directory context
     * @param targetName the target entry name
     * @param attributes the attributes of the entry
     * @throws NamingException error creating entry
     */
    public static void addEntry(final DirContext ctx, final String targetName, final ValueItem<Item>... attributes)
            throws NamingException {

        Attributes attrs = new BasicAttributes(true);
        for (ValueItem<Item> attribute : attributes) {
            String key = attribute.getKey();
            Item value = attribute.getValue();
            if (value instanceof StringItem) {
                StringItem stringItem = ((StringItem) value);
                attrs.put(new BasicAttribute(key, stringItem.getItem()));
            } else if (value instanceof StringArrayItem) {
                StringArrayItem stringArrayItem = ((StringArrayItem) value);
                Attribute subAttrs = new BasicAttribute(key);
                for (String item : stringArrayItem.getItem()) {
                    subAttrs.add(item);
                }
                attrs.put(subAttrs);
            }
        }
        ctx.createSubcontext(targetName, attrs);
    }

    /**
     * Change an existing entry.
     * 
     * @param ctx the directory context
     * @param entry the entry to modify
     * @param values the values to change/add/remove
     * @throws NamingException error changing entry
     */
    public static void modifyMember(final DirContext ctx, final String entry, final ValueItem<Item>... attributes)
            throws NamingException {
        ModificationItem[] mods = new ModificationItem[attributes.length];
        for (int i = 0; i < attributes.length; i++) {
            ValueItem<Item> attribute = attributes[i];
            Item value = attribute.getValue();
            if (value instanceof StringItem) {
                StringItem stringValue = (StringItem) value;
                mods[i] = new ModificationItem(attribute.getMode(), new BasicAttribute(attribute.getKey(),
                        stringValue.getItem()));
            } else if (value instanceof StringArrayItem) {
                StringArrayItem stringArrayValue = (StringArrayItem) value;
                String[] items = stringArrayValue.getItem();
                for (String item : items) {
                    mods[i] = new ModificationItem(attribute.getMode(), new BasicAttribute(attribute.getKey(), item));
                }
            }
        }
        ctx.modifyAttributes(entry, mods);
    }
}
